/**
 * Copyright 2008 Peach Jean Solutions
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.xj4.lifecycle;

import org.xj4.XJ4TestClass;
import org.xj4.util.ServiceLoader;
import org.xj4.spi.TestWrapper;

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.List;
import java.util.ArrayList;

/**
 * Created by IntelliJ IDEA.
 * User: jbunting
 * Date: Nov 9, 2008
 * Time: 9:01:32 PM
 * To change this template use File | Settings | File Templates.
 */
public class LifecycleDescriptor {
  private XJ4TestClass testClass;
  private Field field;
  private ServiceLoader<LifecycleProvider> providers;

  public LifecycleDescriptor(XJ4TestClass testClass, Field field, ServiceLoader<LifecycleProvider> providers) {
    this.testClass = testClass;
    this.field = field;
    this.providers = providers;
  }

  public List<TestWrapper> generateTestWrappers(final Object target) {
    List<TestWrapper> wrappers = new ArrayList<TestWrapper>();
    for(Lifecycle lifecycle: retrieveLifecycles()) {
      wrappers.add(new LifecycleTestWrapper(lifecycle, target));
    }
    return wrappers;
  }

  private class LifecycleTestWrapper implements TestWrapper {
    private Lifecycle lifecycle;
    private Object target;

    private LifecycleTestWrapper(Lifecycle lifecycle, Object target) {
      this.lifecycle = lifecycle;
      this.target = target;
    }

    public void doBefore() throws Throwable {
      lifecycle.start(testClass, field, target);
    }

    public void doAfter() throws Throwable {
      lifecycle.stop(testClass, field, target);
    }
  }

  public boolean isInstance() {
    return !Modifier.isStatic(field.getModifiers());    
  }

  @SuppressWarnings({"ThrowableInstanceNeverThrown"})
  public void validate(List<Throwable> errors) {
    if(!field.isAnnotationPresent(Manage.class)) {
      errors.add(new Exception(String.format("Field %s is not managed!", field.getName())));
      return;
    }
    field.setAccessible(true);
    List<Lifecycle> lifecycles = retrieveLifecycles();
    if(lifecycles.isEmpty()) {
      errors.add(new Exception(String.format("No lifecycle providers could be located for managed field %s.", field.getName())));
    }
    boolean isStatic = Modifier.isStatic(field.getModifiers());
    for(Lifecycle lifecycle : lifecycles) {
      if(isStatic && !lifecycle.supportsClassLevel()) {
        errors.add(new Exception(String.format("Lifecycle type %s does not support static fields.", lifecycle.getClass().getName())));
      }
      if(!isStatic && !lifecycle.supportsInstanceLevel()) {
        errors.add(new Exception(String.format("Lifecycle type %s does not support instance fields.", lifecycle.getClass().getName())));
      }
      if(!lifecycle.requiredType().isAssignableFrom(field.getType())) {
        errors.add(new Exception(String.format("Lifecycle type %s only operates on fields of type %s.", lifecycle.getClass().getName(), lifecycle.requiredType())));
      }
      lifecycle.validate(errors, field);
    }
  }

  protected List<Lifecycle> retrieveLifecycles() {
    Class<? extends Lifecycle>[] lifecycleTypes = getLifecycleClasses();
    List<Lifecycle> lifecycles = new ArrayList<Lifecycle>();
    for(Class<? extends Lifecycle> lifecycleType : lifecycleTypes) {
      try {
        lifecycles.add(lifecycleType.newInstance());
      } catch(Exception ex) {
        throw new IllegalArgumentException("Lifecycle type " + lifecycleType.getName() + "is not instantiable.", ex);
      }
    }
    if(lifecycleTypes.length == 0) {
      LifecycleProvider provider = providers.selectFirst(new LifecycleProviderSelector(field));
      if(provider != null) {
        lifecycles.addAll(provider.handle(field));
      }
    }
    return lifecycles;
  }

  private Class<? extends Lifecycle>[] getLifecycleClasses() {
    Manage manage = field.getAnnotation(Manage.class);
    if(manage == null) {
      return new Class[0];
    }
    Class<? extends Lifecycle>[] lifecycleClasses = manage.lifecycle();
    if(lifecycleClasses.length == 0) {
      lifecycleClasses = manage.value();
    }
    return lifecycleClasses;
  }
}
