/**
 * Copyright 2008 Peach Jean Solutions
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.xj4.lifecycle;

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.List;
import java.util.ArrayList;

import org.xj4.XJ4TestClass;
import org.xj4.util.ServiceLoader;
import org.xj4.spi.TestWrapper;
import org.xj4.spi.MethodWrapperSource;
import org.xj4.spi.ClassWrapperSource;
import org.xj4.spi.TestClassValidator;
import org.apache.commons.logging.LogFactory;
import org.apache.commons.logging.Log;
import org.junit.runners.model.FrameworkMethod;

/**
 * @author jbunting
 * @TODO Document Me!
 * Created by IntelliJ IDEA.
 * Date: Mar 5, 2008
 * Time: 7:41:27 AM
 */
public class LifecycleManager implements ClassWrapperSource, MethodWrapperSource, TestClassValidator {

  private static final Log logger = LogFactory.getLog(LifecycleManager.class);

  private ServiceLoader<LifecycleProvider> providers = new ServiceLoader(LifecycleProvider.class);

  public List<TestWrapper> generateWrapper(XJ4TestClass testClass, FrameworkMethod method, Object target) {
    return generateWrappers(testClass, LifecycleType.INSTANCE, target);
  }

  public List<TestWrapper> generateWrapper(XJ4TestClass testClass) {
    return generateWrappers(testClass, LifecycleType.CLASS, null);
  }

  public void validate(XJ4TestClass testClass, List<Throwable> errors) {
    for(LifecycleDescriptor descriptor: generateDescriptors(testClass)) {
      descriptor.validate(errors);
    }
    for(FrameworkMethod method: testClass.getAnnotatedMethods(PreemptLifecycle.class)) {
      method.validatePublicVoidNoArg(Modifier.isStatic(method.getMethod().getModifiers()), errors);
    }
  }

  protected List<LifecycleDescriptor> generateDescriptors(XJ4TestClass testClass) {
    List<LifecycleDescriptor> descriptors = new ArrayList<LifecycleDescriptor>();
    for(Field field : retrieveManagedFields(testClass)) {
      descriptors.add(new LifecycleDescriptor(testClass, field, providers));
    }
    return descriptors;
  }

  protected List<TestWrapper> generateWrappers(XJ4TestClass testClass, LifecycleType lifecycleType, Object target) {
    List<TestWrapper> wrappers = new ArrayList<TestWrapper>();
    for(LifecycleDescriptor descriptor: generateDescriptors(testClass)) {
      if(lifecycleType.supports(descriptor)) {
        wrappers.addAll(descriptor.generateTestWrappers(target));
      }
    }
    return wrappers;
  }

  protected List<Field> retrieveManagedFields(XJ4TestClass testClass) {
    List<Field> fields = testClass.getAnnotatedFields(Manage.class);
    logger.debug("Found " + fields.size() + " managed fields.");
    return fields;
  }

  protected enum LifecycleType {
    CLASS() {
      public boolean supports(LifecycleDescriptor descriptor) {
        return !descriptor.isInstance();
      }
      public boolean supports(FrameworkMethod frameworkMethod) {
        return Modifier.isStatic(frameworkMethod.getMethod().getModifiers());
      }},
    INSTANCE() {
      public boolean supports(LifecycleDescriptor descriptor) {
        return descriptor.isInstance();
      }
      public boolean supports(FrameworkMethod frameworkMethod) {
        return !Modifier.isStatic(frameworkMethod.getMethod().getModifiers()); 
      }};

    public abstract boolean supports(LifecycleDescriptor descriptor);

    public abstract boolean supports(FrameworkMethod frameworkMethod);
  }

  protected List<TestWrapper> generateCustomTestWrapper(XJ4TestClass testClass, LifecycleType lifecycleType, Object target) {
    List<FrameworkMethod> methods = testClass.getAnnotatedMethods(PreemptLifecycle.class);
    List<TestWrapper> wrappers = new ArrayList<TestWrapper>();
    for(FrameworkMethod method: methods) {
      if(lifecycleType.supports(method)) {
        wrappers.add(new CustomTestWrapper(method, target));
      }
    }
    return wrappers;
  }

  private class CustomTestWrapper implements TestWrapper {
    private FrameworkMethod method;
    private Object target;

    private CustomTestWrapper(FrameworkMethod method, Object target) {
      this.method = method;
      this.target = target;
    }

    public void doBefore() throws Throwable {
      if(valid(PreemptLifecycle.ConditionType.BEFORE)) {
        method.invokeExplosively(target);
      }
    }

    public void doAfter() throws Throwable {
      if(valid(PreemptLifecycle.ConditionType.AFTER)) {
        method.invokeExplosively(target);
      }
    }

    private boolean valid(PreemptLifecycle.ConditionType type) {
      for(PreemptLifecycle.ConditionType configured: method.getAnnotation(PreemptLifecycle.class).value()) {
        if(configured == type) {
          return true;
        }
      }
      return false;
    }
  }
}
